/*---------------------------------------------------
 * file:    topicfast.c
 * purpose: run standard topic model with shortcut sampling
 * usage:   topic2 -help
 * inputs:  T (number of topics), iter (number of iterations), docword.txt
 * outputs: wp.txt, dp.txt, z.txt
 * version: 1.0
 * author:  iporteou@ics.uci.edu, ihler@ics.uci.edu, newman@uci.edu
 * date:    2008
 *-------------------------------------------------*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <assert.h>
#include "topiclib.h"

void updatesort(int n, int *x, int *indx, int *revindx, int ip, int im);

/*------------------------------------------
 * global variables
 *------------------------------------------ */
static int T; // number of topics
static int W; // number of unique words
static int D; // number of docs
static int N; // number of words in corpus

/*==========================================
 * main
 *========================================== */
int main(int argc, char* argv[])
{
  int i=0, iter, ITER, seed=1, t, tt, tt_, tsp, ttt=-1, ttt_chk, t_, Tchk;
  int *w, *d, *z, **wp, **dp, *ztot, *indx, *nsamp, *nsamplda, *d_last_z, *w_last_z, *ztot_last, ztotmin=-1, d_old, w_old, *indx_z, *revindx_z, **indx_dp, **revindx_dp;
  double alpha, beta, Wbeta, *probs, *Ad, *Bw, Z, Zp=-1, Zp_old, currprob, *a, *b, aa=-1, bb=-1, U, sumprobs, sumprobs_old, fac=-1;

  int *ldacounts, *fastcounts;
  int word, doc;
  int *doc_array, *word_array;

  etime();

  if (argc == 1) {
    fprintf(stderr, "usage: %s T iter alpha beta seed\n", argv[0]);
    exit(-1);
  }
  T     = atoi(argv[1]);
  ITER  = atoi(argv[2]);
  alpha = atof(argv[3]);
  beta  = atof(argv[4]);
  seed  = atoi(argv[5]);
  seed  = 2*seed + 1;

  assert(T>0);
  assert(ITER>0);
  assert(alpha>0);
  assert(beta>0);
  assert(seed>0);

  N = countntot("docword.txt");
  w = ivec(N);
  d = ivec(N);
  z = ivec(N);
  read_dw("docword.txt", d, w, &D, &W);

  seedMT(seed);
  Wbeta = W*beta;
  ztot  = ivec(T);
  probs = dvec(T);
  Ad    = dvec(T);
  Bw    = dvec(T);
  a     = dvec(D);
  b     = dvec(W);
  wp    = imat(W,T);
  dp    = imat(D,T);

  indx_z     = ivec(T);
  revindx_z  = ivec(T);
  indx_dp    = imat(D,T);
  revindx_dp = imat(D,T);
  d_last_z   = ivec(D);
  w_last_z   = ivec(W);
  ztot_last  = ivec(W);
  //nsamp    = ivec(N);

  randomassignment(N,T,w,d,z,wp,dp,ztot);
  
  printf("seed  = %d\n", seed);
  printf("N     = %d\n", N);
  printf("W     = %d\n", W);
  printf("D     = %d\n", D);
  printf("T     = %d\n", T);
  printf("ITER  = %d\n", ITER);
  printf("alpha = %f\n", alpha);
  printf("beta  = %f\n", beta);
  printf("fast lda\n");

  // tt  is old topic assignment
  // ttt is new topic assignment
  
  printf("pre iter %.2f sec\n", etime());

  for (iter = 0; iter <= ITER; iter++) {

    for (i = 0; i < N; i++) {

      word       = w[i];
      doc        = d[i];
      doc_array  = dp[doc];
      word_array = wp[word];
      
      tt = z[i];
      word_array[tt]--;
      doc_array[tt]--;
      ztot[tt]--;

      /************************************************************************/
      /************************************************************************/
      /******  ITER 0     ITER 0     ITER 0     ITER 0     ITER 0  ************/
      /************************************************************************/
      /************************************************************************/

      if (iter==0) {
	
	if (i==0) {
	  isort(T, ztot,  -1, indx_z);
	  isort(T, indx_z, 1, revindx_z);
	} else {
	  if (tt != ttt) updatesort(T,ztot,indx_z,revindx_z,ttt,tt);
	}
	ztotmin = ztot[indx_z[T-1]];
	fac     = 1. / (ztotmin + Wbeta);

	// setup indx_dp
	if (i==N-1 || d[i+1]!=d[i]) { // just do last doc
	  isort(T, doc_array,   -1, indx_dp[doc]);    
	  isort(T, indx_dp[doc], 1, revindx_dp[doc]);
	}
	
	// setup norms aa and bb
	for (t = 0; t < T; t++) {
	  Ad[t] = ( doc_array[t] + alpha);
	  Bw[t] = (word_array[t] + beta)  / (ztotmin + Wbeta);
	}
      	a[doc]  = ddot(T,Ad,Ad);
	b[word] = ddot(T,Bw,Bw);
	aa = a[doc];
	bb = b[word] / (fac*fac);

	// usual sample for iter==0
	for (t = 0, Z = 0.0; t < T; t++) {
	  probs[t] = (doc_array[t] + alpha)*(word_array[t] + beta)/(ztot[t] + Wbeta);
	  Z += probs[t];
	}
	U = Z*drand();
	currprob = probs[0];
	ttt = 0;
	while (U > currprob) {
	  ttt++;
	  currprob += probs[ttt];
	}

      } else {
	
	/************************************************************************/
	/************************************************************************/
	/******  ITER > 0     ITER > 0     ITER > 0     ITER > 0  ***************/
	/************************************************************************/
	/************************************************************************/

	d_old = d_last_z[doc];
	w_old = w_last_z[word];

	if (ttt != tt) {
	  updatesort(T,ztot,indx_z,revindx_z,ttt,tt);
	  ztotmin = ztot[indx_z[T-1]];
	  fac     = 1. / (ztotmin + Wbeta);
	}

        if (d_old != tt) {
          a[doc] += 2*(doc_array[d_old]-doc_array[tt]-1);	// equivalent but faster computation
          //a[doc] -= SQR(alpha + doc_array[d_old]-1);
          //a[doc] -= SQR(alpha + doc_array[tt   ]+1);
          //a[doc] += SQR(alpha + doc_array[d_old]  );
          //a[doc] += SQR(alpha + doc_array[tt   ]  );
          updatesort(T,doc_array,indx_dp[doc],revindx_dp[doc],d_old,tt);
        }
        aa = a[doc];

        b[word] *= SQR(ztot_last[word] + Wbeta);
        if (w_old != tt) {
          b[word] += 2*(word_array[w_old]-word_array[tt]-1);	// "" ""
          //b[word] -= SQR(beta + word_array[w_old]-1);
          //b[word] -= SQR(beta + word_array[tt   ]+1);
          //b[word] += SQR(beta + word_array[w_old]  );
          //b[word] += SQR(beta + word_array[tt   ]  );
        }
	bb = b[word];
	b[word] *= (fac*fac);

	U = drand();                                    // draw
	indx = indx_dp[doc];                            // point indx to current sort order
	for (t_ = 0; t_ < T; t_++) {			// for each possible topic (sorted order)
	  t=indx[t_];					// t_ refers to sort #, t "actual" topic #
	  if (t_!=0) probs[t_]=probs[t_-1]; else probs[t_]=0;	// init cumulative probability mass f'n
	  probs[t_] += (doc_array[t] + alpha)*(word_array[t] + beta)/(ztot[t] + Wbeta);	// increment
	  aa -= SQR(alpha + doc_array[t]);			// update current norm-based bounds
	  bb -= SQR(beta  + word_array[t]);
	  if (aa<0) aa=0; if (bb<0) bb=0;			// avoid sqrt(-0) = nan bug
	  Zp_old=Zp; Zp =  probs[t_] + sqrt(aa*bb) * fac;	// save "old" and "new" Z-bounds
	  if (probs[t_] < U*Zp) continue;			// if we haven't found it yet, loop back
	  else {						// otherwise
	    if ( (t_==0) || (U*Zp > probs[t_-1]) ) {      //   if it's this topic, save & quit
	      ttt = indx[t_]; break;
	    } else {                                      //   if it's a previous topic
	      U = (U*Zp_old-probs[t_-1])*Zp/(Zp_old-Zp);  //   compute the shifted & rescaled threshold
	      for (tt_=0; tt_<t_; tt_++) {		  //   then find it in the collection
		if (probs[tt_] >= U) {
		  ttt = indx[tt_];
		  break;                                  //   (break out of tt_ for loop)
		}
	      } break;                                    //   (done finding which prev topic)
	    }
	  }
	}
	//nsamp[i] = tt_;
	
      }
      
      d_last_z[doc]   = ttt;
      w_last_z[word]  = ttt;
      ztot_last[word] = ztotmin;

      z[i] = ttt;
      word_array[ttt]++;
      doc_array[ttt]++;
      ztot[ttt]++;

    }

    printf("iter %d  %.2f sec\n", iter, etime());
    fflush(stdout);

  }

  return 0;
}


void updatesort(int n, int *x, int *indx, int *revindx, int ip, int im) //
{
  int tmp;

  // x: array
  // indx: index array
  // revindx: reverse index array
  // ip: element to be incremented p=plus
  // im: element to be decremented m=minus

  // following has already been performed
  // x[ip]++;
  // x[im]--;
  
  // INCREMENT
  // did ++ get bigger than prev?
  ip = revindx[ip];
  while (ip>0 && x[indx[ip]] > x[indx[ip-1]]) {
    // swap indx
    tmp        = indx[ip];
    indx[ip]   = indx[ip-1];
    indx[ip-1] = tmp;
    // swap revindx
    tmp                 = revindx[indx[ip]];
    revindx[indx[ip]]   = revindx[indx[ip-1]];
    revindx[indx[ip-1]] = tmp;
    ip--;
  }

  // DECREMENT
  // did -- get smaller than next?
  im = revindx[im];
  while (im<n-1 && x[indx[im]] < x[indx[im+1]]) {
    // swap indx
    tmp        = indx[im];
    indx[im]   = indx[im+1];
    indx[im+1] = tmp;
    // swap revindx
    tmp                 = revindx[indx[im]];
    revindx[indx[im]]   = revindx[indx[im+1]];
    revindx[indx[im+1]] = tmp;
    im++;
  }

}
